import argparse
import json
import os
import pandas as pd
import shutil

from blast_ct.nifti.savers import NiftiPatchSaver
from blast_ct.read_config import get_model, get_test_loader
from blast_ct.train import set_device
from blast_ct.trainer.inference import ModelInference, ModelInferenceEnsemble
from blast_ct.utils.model_loader import get_model_path


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def get_remaining_dataset(test_csv_path, prediction_csv_path):
    test_csv = pd.read_csv(test_csv_path, index_col='id')
    prediction_csv = pd.read_csv(prediction_csv_path, index_col='id')
    dataframe_yet_to_run = test_csv.loc[set(test_csv.index) - set(prediction_csv.index)]
    # Saving new dataframe in tmp
    new_test_csv_path = '/tmp/dataframe_yet_to_run.csv'
    dataframe_yet_to_run.to_csv(new_test_csv_path, index_label='id')
    return new_test_csv_path


def run_inference(job_dir, test_csv_path, config_file, device, saved_model_paths, write_prob_maps,
                  do_localisation, save_atlas_and_brain_mask_native_space, num_reg_runs, overwrite, native_space):
    if not os.path.exists(job_dir):
        os.makedirs(job_dir)
        print('Starting new run...')
    else:
        prediction_csv_path = os.path.join(os.path.join(job_dir, 'predictions'), 'prediction.csv')
        if overwrite:
            print('Run already exists, overwriting...')
            shutil.rmtree(job_dir)
            os.makedirs(job_dir)
        elif not overwrite and os.path.exists(prediction_csv_path):
            print('Run already exists, completing run...')
            test_csv_path = get_remaining_dataset(test_csv_path, prediction_csv_path)

    with open(config_file, 'r') as f:
        config = json.load(f)
    model = get_model(config)
    device = set_device(device)
    use_cuda = device.type != 'cpu'
    test_loader = get_test_loader(config, model, test_csv_path, use_cuda)
    extra_output_names = config['test']['extra_output_names'] if 'extra_output_names' in config['test'] else None

    saver = NiftiPatchSaver(job_dir, test_loader, write_prob_maps=write_prob_maps,
                            extra_output_names=extra_output_names, do_localisation=do_localisation,
                            num_reg_runs=num_reg_runs, native_space=native_space,
                            write_registration_info=save_atlas_and_brain_mask_native_space)

    # saver = NiftiPatchSaver(job_dir, test_loader, write_prob_maps=write_prob_maps,
    #                         extra_output_names=extra_output_names, do_localisation=do_localisation,
    #                         num_reg_runs=num_reg_runs, native_space=native_space)
    saved_model_paths = saved_model_paths.split()
    n_models = len(saved_model_paths)
    task = config['data']['task']

    if n_models == 1:
        ModelInference(job_dir, device, model, saver, saved_model_paths[0], task)(test_loader)
    elif n_models > 1:
        ModelInferenceEnsemble(job_dir, device, model, saver, saved_model_paths, task)(test_loader)


def inference():
    install_dir = os.path.dirname(os.path.realpath(__file__))
    default_config = os.path.join(install_dir, 'data/config.json')

    # Build space-separated string for argparse
    default_model_paths = " ".join(
        get_model_path(f"model_{i}.torch_model")
        for i in range(1, 16)
    )

    parser = argparse.ArgumentParser()
    parser.add_argument('--job-dir',
                        required=True,
                        type=str,
                        help='Directory for checkpoints, exports, and '
                             'logs. Use an existing directory to load a '
                             'trained model, or a new directory to retrain')
    parser.add_argument('--test-csv-path',
                        default=None,
                        type=str,
                        help='Path to test csv file with paths of images and masks.')
    parser.add_argument('--config-file',
                        default=default_config,
                        type=str,
                        help='A json configuration file for the job (see example files)')
    parser.add_argument('--device',
                        type=str,
                        default='cpu',
                        help='Device to use for computation')
    parser.add_argument('--saved-model-paths',
                        default=default_model_paths,
                        type=str,
                        help='Path to saved model or list of paths separated by spaces.')
    parser.add_argument('--write-prob-maps',
                        type=str2bool, nargs='?',
                        const=True,
                        default=False)
    parser.add_argument('--do-localisation',
                        type=str2bool, nargs='?',
                        const=True,
                        default=False)
    parser.add_argument('--num-reg-runs',
                        default=1,
                        type=int)
    parser.add_argument('--save-atlas-and-brain-mask-native-space',
                        default=False,
                        action='store_true')
    parser.add_argument('--overwrite',
                        type=str2bool, nargs='?',
                        const=True,
                        default=False)
    parser.add_argument('--native-space',
                        type=str2bool, nargs='?',
                        const=True,
                        default=True)

    parse_args, unknown = parser.parse_known_args()
    run_inference(**parse_args.__dict__)


if __name__ == "__main__":
    inference()



import operator
import os

import SimpleITK as sitk

sitk.ProcessObject_GetGlobalDefaultNumberOfThreads()


class RegistrationToCTTemplate(object):
    def __init__(self, localisation_dir, target_template_path, num_runs=1, debug_mode=False):
        self.target_template_path = target_template_path
        self.target_template = sitk.ReadImage(self.target_template_path)
        self.localisation_dir = localisation_dir
        self.num_runs = num_runs
        self.debug_mode = debug_mode

    def register_image_to_atlas(self, image):
        image = sitk.Cast(sitk.Threshold(image, lower=-1024.0, upper=1e6, outsideValue=-1024.0), sitk.sitkFloat32)
        dimension = self.target_template.GetDimension()

        # Rigid registration
        # Set the initial moving transforms.
        initial_transform_rig = sitk.CenteredTransformInitializer(self.target_template, image,
                                                                  sitk.Euler3DTransform(),
                                                                  sitk.CenteredTransformInitializerFilter.GEOMETRY)

        registration_method_rig = sitk.ImageRegistrationMethod()

        # Similarity metric settings:
        registration_method_rig.SetMetricAsMattesMutualInformation(numberOfHistogramBins=32)

        # Sampling settings:
        registration_method_rig.SetMetricSamplingStrategy(registration_method_rig.REGULAR)
        registration_method_rig.SetMetricSamplingPercentage(0.2)

        # Interpolator settings:
        registration_method_rig.SetInterpolator(sitk.sitkLinear)

        # Optimizer settings:
        registration_method_rig.SetOptimizerAsGradientDescentLineSearch(learningRate=0.1,
                                                                        numberOfIterations=200,
                                                                        convergenceMinimumValue=1e-6,
                                                                        convergenceWindowSize=5)

        # As we are working with a transformation which parameter space includes both translation and rotation, and
        # mm and radians are not commensurate, and we would like the change of on mm and one radian to have a similar
        # effect, we scale these parameters during the optimization:
        registration_method_rig.SetOptimizerScalesFromPhysicalShift()

        # Setup for the multi-resolution framework:
        registration_method_rig.SetShrinkFactorsPerLevel(shrinkFactors=[4, 2])
        registration_method_rig.SetSmoothingSigmasPerLevel(smoothingSigmas=[4, 2])
        registration_method_rig.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()  # mm

        # Set the initial transform

        registration_method_rig.SetInitialTransform(initial_transform_rig)
        final_rig_transform = registration_method_rig.Execute(self.target_template, image)
        iterations_rig = registration_method_rig.GetOptimizerIteration()
        final_metric_value_rig = registration_method_rig.GetMetricValue()

        # Affine Registration
        registration_method_rig.SetMetricAsMattesMutualInformation(numberOfHistogramBins=32)
        registration_method_rig.SetMetricSamplingStrategy(registration_method_rig.REGULAR)
        registration_method_rig.SetMetricSamplingPercentage(0.2)
        registration_method_rig.SetInterpolator(sitk.sitkLinear)
        registration_method_rig.SetOptimizerAsGradientDescentLineSearch(learningRate=0.1,
                                                                        numberOfIterations=200,
                                                                        convergenceMinimumValue=1e-6,
                                                                        convergenceWindowSize=5)
        registration_method_rig.SetOptimizerScalesFromPhysicalShift()
        registration_method_rig.SetShrinkFactorsPerLevel(shrinkFactors=[4, 2])
        registration_method_rig.SetSmoothingSigmasPerLevel(smoothingSigmas=[4, 2])
        registration_method_rig.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()  # mm
        registration_method_rig.SetMovingInitialTransform(final_rig_transform)
        optimized_transform_aff = sitk.AffineTransform(dimension)
        registration_method_rig.SetInitialTransform(optimized_transform_aff, inPlace=True)

        registration_method_rig.Execute(self.target_template, image)
        # SITK 2
        # final_aff_transform = sitk.CompositeTransform([final_rig_transform, optimized_transform_aff])
        # SITK 1
        final_aff_transform = sitk.Transform()
        final_aff_transform.AddTransform(final_rig_transform)
        final_aff_transform.AddTransform(optimized_transform_aff)

        iterations_aff = registration_method_rig.GetOptimizerIteration()
        final_metric_value_aff = registration_method_rig.GetMetricValue()

        min_filter = sitk.MinimumMaximumImageFilter()
        min_filter.Execute(image)

        resampler = sitk.ResampleImageFilter()
        resampler.SetReferenceImage(self.target_template)
        resampler.SetInterpolator(sitk.sitkLinear)
        resampler.SetDefaultPixelValue(min_filter.GetMinimum())
        resampler.SetTransform(final_aff_transform)
        image_resampled_aff = resampler.Execute(image)

        return final_aff_transform, iterations_rig, final_metric_value_rig, iterations_aff, final_metric_value_aff, \
               image_resampled_aff

    def get_best_run(self, final_metric_aff_dict):
        sm_values = {}
        for iteration in range(0, self.num_runs):
            sm_values[iteration] = final_metric_aff_dict.get(iteration, {}).get('final_metric_aff')
        best_iter = min(sm_values.items(), key=operator.itemgetter(1))[0]
        transform, iterations_rig, final_metric_rig, iterations_aff, final_metric_aff, image_resampled_aff = \
            final_metric_aff_dict[best_iter].values()
        return transform, iterations_rig, final_metric_rig, iterations_aff, final_metric_aff, image_resampled_aff

    def __call__(self, data_index, image, image_id):
        final_metric_aff_dict = {}
        for iteration in range(0, self.num_runs):
            try:
                transform, iterations_rig, final_metric_rig, iterations_aff, final_metric_aff, image_resampled_aff = \
                    self.register_image_to_atlas(image)
                final_metric_aff_dict[iteration] = {'transform': transform, 'iterations_rig': iterations_rig,
                                                    'final_metric_rig': final_metric_rig,
                                                    'iterations_aff': iterations_aff,
                                                    'final_metric_aff': final_metric_aff,
                                                    'image_resampled_aff': image_resampled_aff}
            except RuntimeError:
                print(f'Could not register image: {image_id:s}.')
                return None, data_index

        transform, iterations_rig, final_metric_rig, iterations_aff, final_metric_aff, image_resampled_aff \
            = self.get_best_run(final_metric_aff_dict)

        data_index.loc[image_id, 'quality_control_metric'] = final_metric_aff

        if self.debug_mode:
            resampled_image_path = os.path.join(self.localisation_dir, f'{str(image_id):s}_resampled.nii.gz')
            sitk.WriteImage(image_resampled_aff, resampled_image_path)
            transform_path = os.path.join(self.localisation_dir, f'{str(image_id):s}_transform.tfm')
            sitk.WriteTransform(transform, transform_path)

            data_index.loc[image_id, 'iterations_rig'] = iterations_rig
            data_index.loc[image_id, 'final_metric_rig'] = final_metric_rig
            data_index.loc[image_id, 'iterations_aff'] = iterations_aff
            data_index.loc[image_id, 'image_resampled'] = resampled_image_path
            data_index.loc[image_id, 'aff_transform'] = transform_path

        return transform, data_index



import torch
import time
from blast_ct.trainer.logger import get_logger
from blast_ct.trainer.model_trainer import task_predict_fn_dict


class ModelInference(object):
    def __init__(self, job_dir, device, model, saver, saved_model_path, task):
        self.job_dir = job_dir
        self.device = device
        self.model = model
        self.logger = get_logger(job_dir)
        self.saver = saver
        self.saved_model_path = saved_model_path
        self.model.load_state_dict(torch.load(self.saved_model_path, map_location=self.device))
        self.predict_fn = task_predict_fn_dict[task]

    def inference(self, dataloader):
        self.model.eval()
        for inputs in dataloader:
            inputs = {key: value.to(self.device) for key, value in inputs.items()}
            with torch.set_grad_enabled(False):
                logits, state = self.model(**inputs)
            prob, pred = self.predict_fn(logits)
            state.update({'pred': pred, 'prob': prob, 'logits': logits})
            yield state

    def __call__(self, dataloader):
        start_inference = time.time()
        self.model.to(self.device)

        for model_state in self.inference(dataloader):
            message = self.saver(model_state)
            if message is not None:
                self.logger.info(message)

        time_elapsed = time.time() - start_inference
        self.logger.info(
            'Inference completed in {:.0f}h {:.0f}m {:.0f}s'.format(time_elapsed // 3600, time_elapsed % 3600 // 60,
                                                                    time_elapsed % 3600 % 60))


class ModelInferenceEnsemble(ModelInference):
    def __init__(self, job_dir, device, model, saver, saved_model_paths, task):
        super().__init__(job_dir, device, model, saver, saved_model_paths[0], task)
        assert isinstance(saved_model_paths, list)
        self.saved_model_paths = saved_model_paths

    def inference(self, dataloader):
        self.model.eval()
        for inputs in dataloader:
            inputs = {key: value.to(self.device) for key, value in inputs.items()}
            state, prob_sum = {}, None
            for saved_model_path in self.saved_model_paths:
                self.model.load_state_dict(torch.load(saved_model_path, map_location=self.device))
                with torch.set_grad_enabled(False):
                    logits, state = self.model(**inputs)
                probs, preds = self.predict_fn(logits)
                prob_sum = probs if prob_sum is None else prob_sum + probs
            preds = torch.argmax(prob_sum, dim=1)
            probs = prob_sum / len(self.saved_model_paths)
            state.update({'pred': preds, 'prob': probs, 'logits': probs})
            yield state



import os
import logging


def get_logger(job_dir):
    dashes = '-' * 75
    logger = logging.getLogger(job_dir.replace('/', '_'))
    logger.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    log_file_path = os.path.join(job_dir, 'log' + '.txt')
    if not os.path.exists(os.path.dirname(log_file_path)):
        os.makedirs(os.path.dirname(log_file_path))
    fh = logging.FileHandler(log_file_path)
    fh.setLevel(logging.DEBUG)

    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    formatter = logging.Formatter(dashes + job_dir.split('/')[-1] + ': %(asctime)s' + dashes + '\n%(message)s\n')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)
    logger.propagate = False
    return logger



import itertools
import math
import random

import SimpleITK as sitk
import numpy as np
import pandas as pd
import torch
import torch.utils.data as data

from blast_ct.nifti.augmention import RandomAugmentation
from blast_ct.nifti.patch_samplers import PatchSampler
from blast_ct.nifti.rescale import rescale, reorient_image
from blast_ct.nifti.transformation import Transformation


# numpy.random state is discarded at the end of a worker process and does not propagate between workers or to the
# parent process, random state does and hence we need to reseed numpy to ensure each worker process is actually random
def worker_init_fn(worker_id):
    np.random.seed(random.randint(0, 2 ** 32 - 1))


class NiftiDataset(data.Dataset):
    """ nifti dataset for medical imaging segmentation.

    Args:
        data_csv_path: path to csv file containing paths of channels, targets and sampling masks
        target (string): The name of the target column in the csv.
        sampling_mask (string): The name of the sampling mask column in the csv.
        transformation (callable, optional): A list of transformations that are always applied.
        augmentation (callable, optional): A list of random augmentations which are applied with a given probability.
    """

    def __init__(self,
                 data_csv_path,
                 channels,
                 target=None,
                 sampling_mask=None,
                 sample_weight=None,
                 transformation=None,
                 augmentation=None,
                 max_cases_in_memory=0,
                 task='segmentation',
                 resolution=None):
        self.data_index = pd.read_csv(data_csv_path, index_col='id')
        assert task in ['segmentation', 'classification', 'regression']

        self.channels = channels
        self.target = target
        self.sampling_mask = sampling_mask
        self.sample_weight = sample_weight
        self.transformation = [] if transformation is None else transformation
        self.augmentation = [] if augmentation is None else augmentation
        self.max_cases_in_memory = max_cases_in_memory
        self.case_memory = {}
        self.task = task
        self.resolution = resolution

        for el in self.transformation:
            assert isinstance(el, Transformation)
        assert isinstance(self.augmentation, list)
        for el in self.augmentation:
            assert isinstance(el, RandomAugmentation)

    def get_array_from_dataset(self, index, name, is_discrete=False):
        if name in self.data_index:
            image = sitk.ReadImage(self.data_index.loc[index, name])
            image = reorient_image(image, is_discrete)
            if self.resolution is not None:
                image = rescale(self.resolution, image, is_discrete)
            return sitk.GetArrayFromImage(image).astype(np.float32)
        return None

    def get_case_from_disk(self, image_id):
        target = self.get_array_from_dataset(image_id, self.target, is_discrete=True)
        sampling_mask = self.get_array_from_dataset(image_id, self.sampling_mask, is_discrete=True)

        stack = list()
        for channel in self.channels:
            stack.append(self.get_array_from_dataset(image_id, channel, is_discrete=False))
        image = np.stack(stack)

        if sampling_mask is None:
            sampling_mask = np.ones_like(stack[0])

        for transformation in self.transformation:
            image, target, sampling_mask = transformation(image, target, sampling_mask)

        # image level augmentation is only performed once - assuming images are discarded at the end of the epoch (num_workers>0)
        for augmentation in self.augmentation:
            image, target, sampling_mask = augmentation(image, None if self.task == 'classification' else target,
                                                        sampling_mask)
        if self.sample_weight is not None:
            raise NotImplementedError
        sample_weight = None

        # make sure these arrays are read-only so that if they are stored they won't change
        for array in [image, target, sampling_mask, sample_weight]:
            if array is not None:
                array.flags.writeable = False

        return image, target, sampling_mask, sample_weight

    def get_case(self, index):
        image_id = self.data_index.iloc[index].name
        if image_id in self.case_memory:
            image, target, sampling_mask, sample_weight = self.case_memory[image_id]
        else:
            image, target, sampling_mask, sample_weight = self.get_case_from_disk(image_id)
            if self.max_cases_in_memory > 0:
                if len(self.case_memory) >= self.max_cases_in_memory:
                    self.case_memory.pop(list(self.case_memory.keys())[0])
                self.case_memory[image_id] = [image, target, sampling_mask, sample_weight]

        return image, target, sampling_mask, sample_weight

    def to_tensors(self, image, target, sampling_mask=None, sample_weight=None):
        image = torch.tensor(image, dtype=torch.float32)
        target_type = torch.int64 if self.task in ['segmentation', 'classification'] else torch.float64
        nan_tensor = torch.tensor(float('nan'))
        target = torch.tensor(target, dtype=target_type) if target is not None else nan_tensor
        sampling_mask = torch.tensor(sampling_mask, dtype=torch.float32) if sampling_mask is not None else nan_tensor
        sample_weight = torch.tensor(sample_weight, dtype=torch.float32) if sample_weight is not None else nan_tensor
        return {'image': image, 'target': target, 'sampling_mask': sampling_mask, 'sampling_weight': sample_weight}

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target)
        """
        image, target, sampling_mask, sample_weight = self.get_case(index)
        return self.to_tensors(image, target, sampling_mask, sample_weight)

    def __len__(self):
        return len(self.data_index)

    def __repr__(self):
        return 'Nifti Dataset containing {} cases with {} channels.\n'.format(len(self), len(self.channels))


class PatchWiseNiftiDataset(NiftiDataset, data.IterableDataset):
    def __init__(self,
                 patch_sampler,
                 patches_per_image,
                 images_per_epoch,
                 data_csv_path,
                 channels,
                 target=None,
                 sampling_mask=None,
                 sample_weight=None,
                 transformation=None,
                 augmentation=None,
                 max_cases_in_memory=0,
                 sequential=False,
                 resolution=None):

        # save RAM since when num_workers = 0 dataset state is kept
        max_cases_in_memory = images_per_epoch if max_cases_in_memory > images_per_epoch else max_cases_in_memory
        super().__init__(data_csv_path=data_csv_path,
                         channels=channels,
                         target=target,
                         sampling_mask=sampling_mask,
                         sample_weight=sample_weight,
                         transformation=transformation,
                         augmentation=augmentation,
                         max_cases_in_memory=max_cases_in_memory,
                         task='segmentation',
                         resolution=resolution)

        assert isinstance(patch_sampler, PatchSampler)
        self.patch_sampler = patch_sampler
        self.patches_per_image = patches_per_image
        self.images_per_epoch = images_per_epoch
        self.images_in_epoch = None
        self.max_count = None
        self.patch_count = None
        self.sequential = sequential

    def __iter__(self):
        worker_info = data.get_worker_info()
        start = 0
        end = self.images_per_epoch
        if worker_info is not None:  # more than one process data-loading
            per_worker = int(math.ceil((end - start) / float(worker_info.num_workers)))
            start = start + worker_info.id * per_worker
            end = min(start + per_worker, end)
            # save RAM - likely not necessary since dataset state isn't kept when num_workers > 0
            if self.max_cases_in_memory > per_worker:
                self.max_cases_in_memory = per_worker

        # At the start of every epoch sample which images are in this epoch, assuring each worker has the same images
        replace = False if len(self.data_index) >= self.images_per_epoch else True
        self.images_in_epoch = np.random.choice(range(len(self.data_index)), self.images_per_epoch, replace=replace)
        self.images_in_epoch = self.images_in_epoch[start:end]
        self.max_count = len(self.images_in_epoch) * self.patches_per_image
        self.patch_count = 0
        return self

    def __next__(self):
        if self.patch_count >= self.max_count:
            raise StopIteration
        idx = self.patch_count % len(
            self.images_in_epoch) if not self.sequential else self.patch_count // self.patches_per_image
        image_index = self.images_in_epoch[idx]
        image, target, sampling_mask, sample_weight = self.get_case(image_index)
        image_patch, target_patch, sampling_mask_patch = self.patch_sampler(image, target, sampling_mask)
        self.patch_count += 1
        return self.to_tensors(image_patch, target_patch, sampling_mask_patch, sample_weight)

    def __len__(self):
        return int(self.images_per_epoch * self.patches_per_image)


class FullImageToOverlappingPatchesNiftiDataset(NiftiDataset, data.IterableDataset):
    def __init__(self,
                 image_patch_shape,
                 target_patch_shape,
                 data_csv_path,
                 channels,
                 target=None,
                 sampling_mask=None,
                 sample_weight=None,
                 transformation=None,
                 augmentation=None,
                 resolution=None):

        super().__init__(data_csv_path=data_csv_path,
                         channels=channels,
                         target=target,
                         sampling_mask=sampling_mask,
                         sample_weight=sample_weight,
                         transformation=transformation,
                         augmentation=augmentation,
                         max_cases_in_memory=1,
                         task='segmentation',
                         resolution=resolution)
        self.patch_sampler = PatchSampler(image_patch_shape, target_patch_shape)
        self.target_patch_shape = target_patch_shape
        self.index_mapping = []
        self.image_mapping = {}
        for image_index, (image_id, row) in enumerate(self.data_index.iterrows()):
            image = sitk.ReadImage(self.data_index.loc[image_id, self.channels[0]])
            image = reorient_image(image, False)
            if self.resolution is not None:
                image = rescale(self.resolution, image)
            target_shape = image.GetSize()[::-1]
            center_points = self.get_center_points(target_shape, target_patch_shape)
            self.image_mapping[image_index] = (target_shape, center_points)
            for patch_index in range(len(center_points)):
                self.index_mapping.append((image_index, patch_index))
        self.patch_index = None

    @staticmethod
    def get_center_points(shape, patch_shape):
        return list(itertools.product(*((ps * i + ps // 2 for i in range(s // ps + 1)) for s, ps in
                                        zip(shape, patch_shape))))

    def __iter__(self):
        worker_info = data.get_worker_info()
        if worker_info is not None:
            if worker_info.num_workers > 1:
                raise ValueError(
                    'Patches must be sequential for the saver to reconstruct the image hence num_workers must be 0 or 1')
        self.patch_index = 0
        return self

    def __next__(self):
        if self.patch_index >= len(self.index_mapping):
            raise StopIteration
        image_index, patch_index = self.index_mapping[self.patch_index]
        target_shape, center_points = self.image_mapping[image_index]

        image, target, sampling_mask, _ = self.get_case(image_index)

        center = center_points[patch_index]
        image_patch, target_patch, sampling_mask_patch = \
            self.patch_sampler.get_patches(center, image, target, sampling_mask)

        self.patch_index += 1
        return self.to_tensors(image_patch, target_patch, sampling_mask_patch, None)

    def __len__(self):
        return len(self.index_mapping)



import argparse
import json
import os
import shutil
import sys

import pandas as pd

from blast_ct.nifti.savers import NiftiPatchSaver
from blast_ct.read_config import get_model, get_test_loader
from blast_ct.train import set_device
from blast_ct.trainer.inference import ModelInference, ModelInferenceEnsemble
from blast_ct.utils.model_loader import get_model_path   # 🌟 add this


def path(string):
    if os.path.exists(string):
        return string
    else:
        sys.exit(f'File not found: {string}')


def console_tool():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input', metavar='input', type=path, required=True)
    parser.add_argument('--output', metavar='output', type=str, required=True)
    parser.add_argument('--ensemble', action='store_true', default=False)
    parser.add_argument('--device', default='cpu')
    parser.add_argument('--do-localisation', action='store_true', default=False)

    parse_args, unknown = parser.parse_known_args()

    # (validation code here unchanged)

    install_dir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(install_dir, 'data/config.json'), 'r') as f:
        config = json.load(f)

    device = set_device(parse_args.device)
    if device.type == 'cpu':
        config['test']['batch_size'] = 32

    job_dir = '/tmp/blast_ct'
    os.makedirs(job_dir, exist_ok=True)

    test_csv_path = os.path.join(job_dir, 'test.csv')
    pd.DataFrame([['im_0', parse_args.input]], columns=['id', 'image']).to_csv(test_csv_path, index=False)

    model = get_model(config)
    test_loader = get_test_loader(config, model, test_csv_path, use_cuda=(device.type != 'cpu'))
    saver = NiftiPatchSaver(job_dir, test_loader, write_prob_maps=False, do_localisation=parse_args.do_localisation)

    if not parse_args.ensemble:
        model_path = get_model_path("model_1.torch_model")   # 🌟 changed
        ModelInference(job_dir, device, model, saver, model_path, 'segmentation')(test_loader)
    else:
        model_paths = [get_model_path(f"model_{i}.torch_model") for i in range(1, 13)]  # 🌟 changed
        ModelInferenceEnsemble(job_dir, device, model, saver, model_paths, task='segmentation')(test_loader)

    output_dataframe = pd.read_csv(os.path.join(job_dir, 'predictions/prediction.csv'))
    shutil.copyfile(output_dataframe.loc[0, 'prediction'], parse_args.output)
    shutil.rmtree(job_dir)



import numpy as np

from blast_ct.nifti.augmention import RandomAugmentation


# All patch samplers can't alter state after __init__ because of dataset class when num_workers > 0

def get_patch_and_padding(shape, patch_size, center):
    padding = [[0, 0], [0, 0], [0, 0]]
    patch = []
    for dim, (size, ps, c) in enumerate(zip(shape, patch_size, center)):
        start = c - ps // 2
        end = start + ps
        if start < 0:
            padding[dim][0] = -start
            start = 0
        if end > size:
            padding[dim][1] = end - size
            end = size
        patch.append(slice(start, end, 1))

    return patch, padding


class PatchSampler(object):
    def __init__(self, image_patch_shape, target_patch_shape, augmentation=None):
        if any((i < l) for i, l in zip(image_patch_shape, target_patch_shape)):
            raise ValueError('Label map patch size must be smaller or equal to image patch size.')
        self.image_patch_size = image_patch_shape
        self.target_patch_size = target_patch_shape
        self.augmentation = [] if augmentation is None else augmentation
        assert isinstance(self.augmentation, list)
        for el in self.augmentation:
            assert isinstance(el, RandomAugmentation)

    def get_target_patch(self, target, center):
        if target is None:
            return target
        patch, padding = get_patch_and_padding(target.shape, self.target_patch_size, center)
        target_patch = target[tuple(patch)]
        if any([any(p) for p in padding]):
            target_patch = np.pad(target_patch, padding, mode='constant', constant_values=0)
        return target_patch

    def get_image_patch(self, image, center):
        image_shape = image.shape
        patch, padding = get_patch_and_padding(image_shape[1:], self.image_patch_size, center)
        padding = [[0, 0]] + padding
        patch = [slice(0, image_shape[0], 1)] + patch

        image_patch = image[tuple(patch)]

        if any([any(p) for p in padding]):
            image_patch = np.pad(image_patch, padding, mode='edge')
        return image_patch

    def get_patches(self, center, image, target, mask):
        image_patch = self.get_image_patch(image, center)
        target_patch = self.get_target_patch(target, center)
        mask_patch = self.get_target_patch(mask, center)
        return image_patch, target_patch, mask_patch

    def sample_patch_center(self, target, mask):
        raise NotImplementedError

    def __call__(self, image, target, mask=None):
        # ensure that the mask has at least one point
        mask = np.ones_like(target) if mask is None else mask
        if np.sum(mask) == 0:
            raise ValueError('Empty sampling mask')

        center = self.sample_patch_center(target, mask)
        image_patch, target_patch, mask_patch = self.get_patches(center, image, target, mask)
        for augmentation in self.augmentation:
            image_patch, target_patch, mask_patch = augmentation(image_patch, target_patch, mask_patch)
        # these copies are only here because of flipping and should be removed in the future
        return image_patch.copy(), target_patch.copy(), mask_patch.copy()


class StochasticPatchSampler(PatchSampler):
    def get_sampling_mask(self, target, mask):
        raise NotImplementedError

    def sample_patch_center(self, target, mask):
        sampling_mask = self.get_sampling_mask(target, mask)
        points = np.argwhere(sampling_mask)
        center = points[np.random.choice(len(points))]
        return center


class RandomPatchSampler(StochasticPatchSampler):
    def __init__(self, image_patch_shape, target_patch_shape, augmentation=None):
        super().__init__(image_patch_shape, target_patch_shape, augmentation)

    def get_sampling_mask(self, target, mask):
        return mask


class ConditionalPatchSampler(StochasticPatchSampler):
    def __init__(self, image_patch_shape, target_patch_shape,
                 class_probabilities, augmentation=None, n_tries=3):
        super().__init__(image_patch_shape, target_patch_shape, augmentation)
        self.class_probabilities = class_probabilities
        self.n_tries = n_tries

    def get_sampling_mask(self, target, mask):
        # try sampling points from a random label a maximum of n_tries
        for i in range(self.n_tries):
            label = np.random.choice(len(self.class_probabilities), p=self.class_probabilities)
            sampling_mask = np.logical_and(target == label, mask)
            if sampling_mask.any():
                return sampling_mask
        # if sampling mask is empty default to mask
        return mask


class ForegroundBackgroundPatchSampler(StochasticPatchSampler):
    def __init__(self, image_patch_shape, target_patch_shape,
                 foreground_probability=.5, augmentation=None, n_tries=3):
        super().__init__(image_patch_shape, target_patch_shape, augmentation)
        self.foreground_probability = foreground_probability
        self.n_tries = n_tries

    def get_sampling_mask(self, target, mask):
        # try sampling points from a random label a maximum of n_tries
        for i in range(self.n_tries):
            is_foreground = np.random.choice((False, True),
                                             p=(1 - self.foreground_probability, self.foreground_probability))
            if is_foreground:
                sampling_mask = np.logical_and(target != 0, mask)
            else:
                sampling_mask = np.logical_and(target == 0, mask)
            if sampling_mask.any():
                return sampling_mask
        # if sampling mask is empty default to mask
        return mask


class BoundingBoxCenteredPatchSampler(PatchSampler):
    def __init__(self, image_patch_shape, target_patch_shape, augmentation):
        super().__init__(image_patch_shape, target_patch_shape, augmentation)

    def sample_patch_center(self, target, mask):
        return np.array(tuple((np.max(arr) - np.min(arr)) // 2 for arr in np.where(mask)))



import numpy as np


# All transformations can't alter state after __init__ because of dataset class when num_workers > 0


class Transformation(object):
    """
    Abstract class for case transformation.
    __call__: When called a CaseTransformation should return an image and target with the same shape as the input.
    """

    def __call__(self, image, target, sampling_mask):
        raise NotImplementedError


class IntensityWindowNormalization(Transformation):
    def __init__(self, lower_bound, upper_bound, map_upper_bound_to_lower_bound=False):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.map_upper_bound_to_lower_bound = map_upper_bound_to_lower_bound

    def __call__(self, image, target, sampling_mask):

        sampling_mask = np.logical_and(sampling_mask,
                                       np.squeeze(np.logical_and(image >= self.lower_bound, image <= self.upper_bound)))

        image[image <= self.lower_bound] = self.lower_bound
        if self.map_upper_bound_to_lower_bound:
            image[image >= self.upper_bound] = self.lower_bound
        else:
            image[image >= self.upper_bound] = self.upper_bound

        # normalize
        image = (2. * image - self.lower_bound - self.upper_bound) / (self.upper_bound - self.lower_bound)

        return image, target, sampling_mask


class MaskImageUsingSamplingMask(Transformation):
    def __init__(self, outside_value):
        self.outside_value = outside_value

    def __call__(self, image, target, sampling_mask):
        image[np.broadcast_to(np.logical_not(sampling_mask), shape=image.shape)] = self.outside_value
        return image, target, sampling_mask



import os
import pandas as pd
import SimpleITK as sitk
import numpy as np
import torch
import pandas as pd
from blast_ct.localisation.localise_lesions import LesionVolumeLocalisationMNI
from blast_ct.localisation.register_to_template import RegistrationToCTTemplate
from blast_ct.nifti.datasets import FullImageToOverlappingPatchesNiftiDataset
from blast_ct.nifti.patch_samplers import get_patch_and_padding
from blast_ct.nifti.rescale import create_reference_reoriented_image

CLASS_NAMES = ['background', 'iph', 'eah', 'oedema', 'ivh']


def add_predicted_volumes_to_dataframe(dataframe, id_, array, resolution):
    voxel_volume_ml = np.prod(resolution) / 1000.
    for i, class_name in enumerate(CLASS_NAMES):
        if i == 0:
            continue
        volume = np.sum(array == i) * voxel_volume_ml
        dataframe.loc[id_, f'{class_name:s}_predicted_volume_ml'] = round(volume, 2)
    return dataframe


def save_image(output_array, input_image, path, resolution):
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    image = sitk.GetImageFromArray(output_array)
    reference = create_reference_reoriented_image(input_image)
    image.SetOrigin(reference.GetOrigin())
    image.SetDirection(reference.GetDirection())
    image.SetSpacing(resolution)
    image = sitk.Resample(image, input_image, sitk.Transform(), sitk.sitkNearestNeighbor, 0)
    sitk.WriteImage(image, path)
    return image


def get_num_maps(patches):
    shape = patches[0].shape
    if len(shape) == 3:
        return 1
    elif len(shape) == 4:
        return shape[0]
    else:
        raise ValueError('Trying to save a tensor with dimensionality which is not 3 or 4.')


def reconstruct_image(patches, image_shape, center_points, patch_shape):
    num_maps = get_num_maps(patches)
    assert len(patches) == len(center_points)
    padded_shape = tuple(s - s % ps + ps for s, ps in zip(image_shape, patch_shape))
    reconstruction = np.zeros(shape=(num_maps,) + padded_shape)
    for center, patch in zip(center_points, patches):
        slices, _ = get_patch_and_padding(padded_shape, patch_shape, center)
        reconstruction[(slice(0, num_maps, 1),) + tuple(slices)] = patch
    reconstruction = reconstruction[(slice(0, num_maps, 1),) + tuple(slice(0, s, 1) for s in image_shape)]
    reconstruction = reconstruction.transpose(tuple(range(1, reconstruction.ndim)) + (0,))
    return reconstruction


class Localisation(object):
    def __init__(self, localisation_dir, num_runs, native_space, write_registration_info):
        if not os.path.exists(localisation_dir):
            os.makedirs(localisation_dir)
        self.localisation_dir = localisation_dir

        asset_dir = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
                                 'data/localisation_files')
        target_template_path = os.path.join(asset_dir, 'ct_template.nii.gz')
        atlas_label_map_path = os.path.join(asset_dir, 'atlas_template_space.nii.gz')
        brain_mask_path = os.path.join(asset_dir, 'ct_template_mask.nii.gz')
        roi_dictionary_csv = os.path.join(asset_dir, 'atlas_labels.csv')

        self.register = RegistrationToCTTemplate(localisation_dir, target_template_path, num_runs=num_runs)
        self.localise = LesionVolumeLocalisationMNI(localisation_dir, native_space, atlas_label_map_path,
                                                    brain_mask_path, roi_dictionary_csv, 'prediction',
                                                    write_registration_info)

    def __call__(self, data_index, image_id, input_image, prediction):
        transform, data_index = self.register(data_index, input_image, image_id)
        if transform is not None:
            return self.localise(transform, data_index, image_id, prediction)
        return data_index


class NiftiPatchSaver(object):
    def __init__(self, job_dir, dataloader, write_prob_maps=True, extra_output_names=None, do_localisation=False,
                 num_reg_runs=1, native_space=True, write_registration_info=False):
        assert isinstance(dataloader.dataset, FullImageToOverlappingPatchesNiftiDataset)

        self.prediction_dir = os.path.join(job_dir, 'predictions')
        self.prediction_csv_path = os.path.join(self.prediction_dir, 'prediction.csv')
        self.dataloader = dataloader
        self.dataset = dataloader.dataset
        self.write_prob_maps = write_prob_maps
        self.patches = []
        self.extra_output_patches = {key: [] for key in extra_output_names} if extra_output_names is not None else {}
        self.image_index = 0
        if os.path.exists(self.prediction_csv_path):
            self.prediction_index = pd.read_csv(self.prediction_csv_path, index_col='id')
        else:
            self.prediction_index = pd.DataFrame(columns=self.dataset.data_index.columns)

        localisation_dir = os.path.join(job_dir, 'localisation')
        self.localisation = Localisation(localisation_dir, num_reg_runs, native_space, write_registration_info) if do_localisation else None

    def reset(self):
        self.image_index = 0
        self.patches = []
        if self.extra_output_patches is not None:
            self.extra_output_patches = {key: [] for key in self.extra_output_patches}

    def append(self, state):
        if self.write_prob_maps:
            self.patches += list(state['prob'].cpu().detach())
        else:
            self.patches += list(state['pred'].cpu().detach())
        for name in self.extra_output_patches:
            self.extra_output_patches[name] += list(state[name].cpu().detach())

    def __call__(self, state):
        self.append(state)
        target_shape, center_points = self.dataset.image_mapping[self.image_index]
        target_patch_shape = self.dataset.patch_sampler.target_patch_size
        patches_in_image = len(center_points)

        if len(self.patches) >= patches_in_image:
            to_write = {}
            image_id = self.dataset.data_index.iloc[self.image_index].name
            input_image = sitk.ReadImage(self.dataset.data_index.loc[image_id, self.dataset.channels[0]])
            patches = list(torch.stack(self.patches[0:patches_in_image]).numpy())
            self.patches = self.patches[patches_in_image:]
            reconstruction = reconstruct_image(patches, target_shape, center_points, target_patch_shape)

            if self.write_prob_maps:
                to_write['prob_maps'] = reconstruction
                to_write['prediction'] = np.argmax(reconstruction, axis=-1).astype(np.float64)
            else:
                to_write['prediction'] = reconstruction

            for name in self.extra_output_patches:
                patches = list(torch.stack(self.extra_output_patches[name][0:patches_in_image]).numpy())
                self.extra_output_patches[name] = self.extra_output_patches[name][patches_in_image:]
                images = reconstruct_image(patches, target_shape, center_points, target_patch_shape)
                to_write[name] = images

            resolution = self.dataset.resolution if self.dataset.resolution is not None else input_image.GetSpacing()
            columns = self.dataset.data_index.columns
            self.prediction_index.loc[image_id, columns] = self.dataset.data_index.loc[image_id, columns]
            for name, array in to_write.items():
                path = os.path.join(self.prediction_dir, f'{str(image_id):s}_{name:s}.nii.gz')
                self.prediction_index.loc[image_id, name] = path
                try:
                    output_image = save_image(array, input_image, path, resolution)
                    if name == 'prediction':
                        self.prediction_index = add_predicted_volumes_to_dataframe(self.prediction_index, image_id,
                                                                                   sitk.GetArrayFromImage(output_image),
                                                                                   output_image.GetSpacing())
                        if self.localisation is not None:
                            self.prediction_index = self.localisation(self.prediction_index, image_id, input_image,
                                                                      output_image)
                    message = f"{self.image_index:d}/{len(self.dataset.data_index):d}: Saved prediction for {str(image_id)}."
                except:
                    message = f"{self.image_index:d}/{len(self.dataset.data_index):d}: Error saving prediction for {str(image_id)}."
                    continue

            self.prediction_index.to_csv(self.prediction_csv_path, index_label='id')

            self.image_index += 1
            if self.image_index >= len(self.dataset.image_mapping):
                self.reset()

            return message

        return None



from abc import ABC
import numpy as np
import torch


def to_np_cpu(tensor):
    return tensor.to('cpu').detach().numpy()


def report_scalar(name, scalar):
    return '{0}: {1:.4f}\t'.format(name, scalar).ljust(20)


def report_mean_and_std(name, array):
    return '{}: {:.4f} ± {:.4f}\t'.format(name, np.mean(array), np.std(array))


class Metric(ABC):
    def __init__(self, initial_value):
        self.running_value = initial_value
        self.value = None

    def increment(self, model_state):
        raise NotImplementedError

    def save_and_reset(self):
        raise NotImplementedError

    def report(self):
        raise NotImplementedError

    def log_to_tensorboard(self, epoch, writer, tag):
        raise NotImplementedError


class Loss(Metric):
    def __init__(self, device):
        Metric.__init__(self, torch.zeros(1, device=device))

    def increment(self, model_state):
        self.running_value += model_state['loss']

    def save_and_reset(self):
        self.value = float(self.running_value.to('cpu').detach())
        self.running_value[:] = 0

    def report(self):
        return report_scalar('loss', self.value) + '\n'

    def log_to_tensorboard(self, epoch, writer, tag):
        writer.add_scalar(tag, self.value, global_step=epoch)


class TrackLogitDistribution(Metric):
    # assumes shape (batch_size, num_maps, ...)
    def __init__(self, name, num_maps: int):
        self.name = name
        Metric.__init__(self, [[0, 0, 0] for _ in range(num_maps)])
        self.num_maps = num_maps
        self.value = [None for _ in range(num_maps)]

    def increment(self, model_state):
        tensor = model_state[self.name].detach()
        for i in range(self.num_maps):
            t = tensor[:, i]
            self.running_value[i][0] += t.numel()
            self.running_value[i][1] += torch.sum(t)
            self.running_value[i][2] += torch.sum(t * t)

    def save_and_reset(self):
        for i in range(self.num_maps):
            n = self.running_value[i][0]
            sum_ = self.running_value[i][1]
            sum_square = self.running_value[i][2]
            var = (sum_square - ((sum_ ** 2) / n)) / (n - 1)
            mean = sum_ / n
            self.value[i] = [mean.cpu().numpy(), torch.sqrt(var).cpu().numpy()]
        self.running_value = [[0, 0, 0] for _ in range(self.num_maps)]

    def report(self):
        messages = []
        for i, value in enumerate(self.value):
            messages.append(f'{self.name:s}_{i:d}: {value[0]:.4f} ± {value[1]:.4f}')
        l = max([len(m) for m in messages])
        message = ''
        for m in messages:
            message += m.ljust(l) + '\t'
        message += '\n'
        return message

    def log_to_tensorboard(self, epoch, writer, tag):
        for i, value in enumerate(self.value):
            writer.add_scalar(f'{tag:s}_{i:d}/mean', value[0], global_step=epoch)
            writer.add_scalar(f'{tag:s}_{i:d}/stddev', value[1], global_step=epoch)


class TrackTensor(Metric):

    def __init__(self, name):
        self.name = name
        Metric.__init__(self, [])

    def increment(self, model_state):
        self.value += list(model_state[self.name])

    def save_and_reset(self):
        self.value = to_np_cpu(torch.squeeze(torch.stack(tuple(self.running_value))))
        self.running_value = []

    def report(self):
        return report_mean_and_std(self.name, self.value)

    def log_to_tensorboard(self, epoch, writer, tag):
        pass


class RunningConfusionMatrix(Metric):

    def __init__(self, num_classes, device):
        super().__init__(torch.zeros((num_classes, num_classes), device=device))
        self.eye = torch.eye(num_classes, num_classes, device=device)

    def compute_confusion_matrix(self, labels, preds):
        return torch.einsum('nd,ne->de', self.eye[labels.flatten()], self.eye[preds.flatten()])

    def increment(self, model_state):
        self.running_value += self.compute_confusion_matrix(model_state['target'], model_state['pred'])

    def save_and_reset(self):
        self.value = to_np_cpu(self.running_value)
        self.running_value[:] = 0

    def report(self):
        return ''

    def log_to_tensorboard(self, epoch, writer, tag):
        pass


def calc_accuracy(cm):
    return np.repeat(np.sum(np.diag(cm)) / np.sum(cm), cm.shape[0])


def calc_precision(cm):
    return np.diag(cm) / np.sum(cm, axis=0)


def calc_recall(cm):
    return np.diag(cm) / np.sum(cm, axis=1)


def calc_f1_score(cm):
    return np.diag(cm) / np.sum(cm, axis=1)


class ClassificationMetrics(RunningConfusionMatrix):
    metric_fns = {'accuracy': calc_accuracy,
                  'precision': calc_precision,
                  'recall': calc_recall,
                  'f1_score': calc_f1_score}

    def __init__(self, device, class_names):

        self.class_names = class_names

        RunningConfusionMatrix.__init__(self, len(class_names), device)
        self.metrics = dict.fromkeys(self.metric_fns.keys())

    def save_and_reset(self):
        super().save_and_reset()
        cm = self.value
        for metric in self.metrics:
            self.metrics[metric] = self.metric_fns[metric](cm)

    def log_to_tensorboard(self, epoch, writer, tag):
        for i, class_name in enumerate(self.class_names):
            for metric, value in self.metrics.items():
                writer.add_scalar(class_name + '/' + str(metric), value[i], global_step=epoch)

    def report(self):
        message = ''
        for i, class_name in enumerate(self.class_names):
            message += class_name.upper().ljust(20) + ':\t'
            for metric, value in self.metrics.items():
                message += report_scalar(metric, value[i])
            message += '\n'
        return message


class SegmentationMetrics(ClassificationMetrics):
    def __init__(self, device, class_names):
        class_names[0] = 'Foreground'
        super().__init__(device, class_names)

    @staticmethod
    def merge_cm_classes(cm):
        new_cm = np.zeros(shape=(2, 2))
        fi = list(range(1, cm.shape[0]))
        new_cm[1, 1] = np.sum(cm[fi, fi])
        new_cm[0, 0] = cm[0, 0]
        new_cm[1, 0] = np.sum(cm[fi, 0])
        new_cm[0, 1] = np.sum(cm[0, fi])
        return new_cm

    def save_and_reset(self):
        super().save_and_reset()
        cm = self.value
        f_cm = self.merge_cm_classes(cm)
        for metric in self.metrics:
            self.metrics[metric][0] = self.metric_fns[metric](f_cm)[1]



import torch
import torch.nn as nn


class CrossEntropyLoss(nn.CrossEntropyLoss):
    def __init__(self, weight=None, size_average=None, ignore_index=-100, reduce=None, reduction='mean'):
        super().__init__(weight, size_average, ignore_index, reduce, reduction)

    def forward(self, logits: torch.tensor, target: torch.tensor, **kwargs):
        return super().forward(logits, target)



from abc import ABC

import torch.nn as nn
import torch.nn.functional as F


def calculate_convolution_output_size(input_size, conv):
    assert (isinstance(conv, nn.Conv3d) or isinstance(conv, nn.Conv2d))
    output_size = tuple((i + 2 * p - k - (k - 1) * (d - 1)) // s + 1 for i, p, k, d, s
                        in zip(input_size, conv.padding, conv.kernel_size, conv.dilation, conv.stride))
    return output_size


def calculate_convolution_input_size(output_size, conv):
    assert (isinstance(conv, nn.Conv3d) or isinstance(conv, nn.Conv2d))
    input_size = tuple((o - 1) * s + (k - 1) * (d - 1) + k - 2 * p for o, p, k, d, s
                       in zip(output_size, conv.padding, conv.kernel_size, conv.dilation, conv.stride))
    return input_size


def crop_center(x, size):
    if tuple(x.shape[2:]) == size:
        return x
    crop = tuple((slice(0, x.shape[0], 1), slice(0, x.shape[1], 1)))
    crop += tuple(slice(c // 2 - s // 2, c // 2 + s // 2 + s % 2, 1) for c, s in zip(x.shape[2:], size))
    return x[crop]


class BiomedicalModule(nn.Module, ABC):
    def __init__(self):
        super().__init__()

    def calculate_input_size(self, output_size):
        raise NotImplementedError

    def calculate_output_size(self, input_size):
        raise NotImplementedError

    def update_fov_and_scale_factor(self, fov, scale_factor):
        raise NotImplementedError


class BiomedicalBlock(BiomedicalModule, ABC):
    def __init__(self):
        super().__init__()
        self.output_sizes = {}
        self.forward_layers = None

    def get_output_size(self, input_size):
        if input_size not in self.output_sizes:
            output_size = self.calculate_output_size(input_size)
            assert all(o > 0 for o in output_size)

            correct_input_size = self.calculate_input_size(output_size)
            if input_size != correct_input_size:
                raise ValueError('The input size is not allowed because it would be waste memory.'
                                 ' Use {} instead for an output size of {}'.format(str(correct_input_size),
                                                                                   str(output_size)))
            self.output_sizes[input_size] = output_size
        return self.output_sizes[input_size]

    #  reversed(self.path._modules.values()):
    def calculate_input_size(self, output_size):
        input_size = output_size
        for layer in reversed(self.layers):
            input_size = layer.calculate_input_size(input_size)
        return input_size

    # note that the calc size function in not invertible due to the downsampling and upsamping
    def calculate_output_size(self, input_size):
        output_size = input_size
        for layer in self.layers:
            output_size = layer.calculate_output_size(output_size)
        return output_size

    def update_fov_and_scale_factor(self, fov, scale_factor):
        for layer in self.layers:
            fov, scale_factor = layer.update_fov_and_scale_factor(fov, scale_factor)
        return fov, scale_factor

    def calculate_fov(self):
        # input_size = (10000,10000,10000)
        # return tuple(i - o for i,o in zip(input_size, self.calculate_output_size(input_size)))
        fov, _ = self.update_fov_and_scale_factor((1, 1, 1), (1, 1, 1))
        return fov


class SqueezeAndExciteBlock(nn.Module):
    def __init__(self, in_planes, dim=2):
        super(SqueezeAndExciteBlock, self).__init__()
        self.dim = dim
        if self.dim == 2:
            self.fc1 = nn.Conv2d(in_planes, in_planes // 2, kernel_size=1)
            self.fc2 = nn.Conv2d(in_planes // 2, in_planes, kernel_size=1)
        if self.dim == 3:
            self.fc1 = nn.Conv3d(in_planes, in_planes // 2, kernel_size=1)
            self.fc2 = nn.Conv3d(in_planes // 2, in_planes, kernel_size=1)
        else:
            raise ValueError('The spatial dimensionality of the kernel ({0:d}) is not supported.'.format(self.dim))

    def forward(self, x):
        if self.dim == 2:
            w = F.avg_pool2d(x, x.size(2))
        elif self.dim == 3:
            w = F.avg_pool3d(x, x.size(2))
        else:
            raise ValueError('The spatial dimensionality of the kernel ({0:d}) is not supported.'.format(self.dim))
        w = F.relu(self.fc1(w))
        w = self.fc2(w).sigmoid()
        out = x * w
        return out


class PreActBlock(BiomedicalModule):
    """Pre-activation version of the BasicBlock."""

    def __init__(self, in_planes, planes, kernel_size=(3, 3, 3), stride=(1, 1, 1), dropout_prob=0.0, se=False):
        super(PreActBlock, self).__init__()
        self.dim = len(kernel_size)  # 2D or 3D
        assert len(stride) == self.dim
        if self.dim == 2:
            self.bn = nn.BatchNorm2d(in_planes)
            self.conv = nn.Conv2d(in_planes, planes, kernel_size=kernel_size, stride=stride, bias=False)
        elif self.dim == 3:
            self.bn = nn.BatchNorm3d(in_planes)
            self.conv = nn.Conv3d(in_planes, planes, kernel_size=kernel_size, stride=stride, bias=False)
        else:
            raise ValueError('The spatial dimensionality of the kernel ({0:d}) is not supported.'.format(self.dim))

        self.dropout = nn.Dropout(dropout_prob)
        self.prelu = nn.PReLU()
        self.shortcut_slices = 2 * (slice(0, None, 1),) + \
                               tuple(slice(k // 2, None if k == 1 else (-k // 2 + 1), s) for k, s in
                                     zip(kernel_size, stride))

        self.se = SqueezeAndExciteBlock(planes, self.dim) if se else None

    def calculate_input_size(self, output_size):
        return calculate_convolution_input_size(output_size, self.conv)

    def calculate_output_size(self, input_size):
        return calculate_convolution_output_size(input_size, self.conv)

    def forward(self, x):
        out = self.prelu(self.dropout(self.bn(x)))
        out = self.conv(out)
        out = self.se(out) if self.se is not None else out
        if x.shape[1] == out.shape[1]:
            out += x[self.shortcut_slices]

        return out

    def update_fov_and_scale_factor(self, fov, scale_factor):
        fov = tuple(f + s * (k - 1) for f, s, k in zip(fov, scale_factor, self.conv.kernel_size))
        return fov, scale_factor


class DownSample(BiomedicalModule):

    def __init__(self, scale_factor):
        super().__init__()
        self.scale_factor = scale_factor
        self.slices = 2 * (slice(0, None, 1),) + tuple(slice(s // 2, None, s) for s in self.scale_factor)

    def calculate_input_size(self, output_size):
        return tuple(o * s for o, s in zip(output_size, self.scale_factor))

    def calculate_output_size(self, input_size):
        return tuple((i - s) // s + 1 for i, s in zip(input_size, self.scale_factor))

    def forward(self, x):
        return x[self.slices]

    def update_fov_and_scale_factor(self, fov, scale_factor):
        # fov = tuple(f * s for f, s in zip(fov, self.scale_factor)) # under debate
        scale_factor = tuple(s0 * s1 for s0, s1 in zip(scale_factor, self.scale_factor))
        return fov, scale_factor


class UpSample(BiomedicalModule):
    def __init__(self, scale_factor):
        super().__init__()
        self.scale_factor = scale_factor

    def calculate_input_size(self, output_size):
        return tuple(o // s + bool(o % s) for o, s in zip(output_size, self.scale_factor))

    def calculate_output_size(self, input_size):
        return tuple(i * s for i, s in zip(input_size, self.scale_factor))

    def forward(self, x):
        return nn.functional.interpolate(x, scale_factor=self.scale_factor, mode='nearest')

    def update_fov_and_scale_factor(self, fov, scale_factor):
        scale_factor = tuple(s0 / float(s1) for s0, s1 in zip(scale_factor, self.scale_factor))
        return fov, scale_factor


class UpConv(BiomedicalModule):
    def __init__(self, in_channels, out_channels, scale_factor):
        super().__init__()
        self.scale_factor = scale_factor
        padding = ()
        for k in scale_factor:
            padding += ((k - 1) // 2 + bool((k - 1) % 2), (k - 1) // 2)

        dim = len(scale_factor)  # 2D or 3D
        if dim == 2:
            self.padding = nn.ReplicationPad2d(padding)
            self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=scale_factor)
        elif dim == 3:
            self.padding = nn.ReplicationPad3d(padding)
            self.conv = nn.Conv3d(in_channels, out_channels, kernel_size=scale_factor)
        else:
            raise ValueError('The spatial dimensionality of the scale factor ({0:d}) is not supported.'.format(dim))

    def calculate_input_size(self, output_size):
        return tuple(o // s + bool(o % s) for o, s in zip(output_size, self.scale_factor))

    def calculate_output_size(self, input_size):
        return tuple(i * s for i, s in zip(input_size, self.scale_factor))

    def forward(self, x):
        out = self.conv(self.padding(nn.functional.interpolate(x, scale_factor=self.scale_factor, mode='nearest')))
        assert all([i == j * 2 for i, j in zip(out.shape[2:], x.shape[2:])])
        return out

    def update_fov_and_scale_factor(self, fov, scale_factor):
        scale_factor = tuple(s0 / float(s1) for s0, s1 in zip(scale_factor, self.scale_factor))
        return fov, scale_factor


import argparse
import json
import os
import random

import numpy as np
import torch
import torch.nn

from blast_ct.read_config import get_model, get_optimizer, get_loss, get_train_loader, get_valid_loader, \
    get_test_loader, get_training_hooks
from blast_ct.trainer.model_trainer import ModelTrainer
from blast_ct.utils.config_loader import get_config


def run_job(job_dir, train_csv_path, valid_csv_path, config_file,
            num_epochs, device, random_seed, model_index=None):
    with open(config_file, 'r') as f:
        config = json.load(f)
    if not os.path.exists(job_dir):
        os.makedirs(job_dir)
    set_random_seed(random_seed)
    device = set_device(device)
    print('Setting up configuration...')
    task = config['data']['task']
    model = get_model(config)
    use_cuda = device.type != 'cpu'
    train_loader = get_train_loader(config, model, train_csv_path, use_cuda)
    valid_loader = get_valid_loader(config, model, valid_csv_path, use_cuda)
    test_loader = get_test_loader(config, model, valid_csv_path, use_cuda)
    optimizer = get_optimizer(config, model)
    criterion = get_loss(config)
    hooks = get_training_hooks(job_dir, config, device, valid_loader, test_loader)
    # Override naming: save as model_{index}.torch_model instead of model_{epoch}.torch_model
    if model_index is not None:
        for h in hooks:
            from blast_ct.trainer.hooks import ModelSaverHook
            if isinstance(h, ModelSaverHook):
                h.filename = f"model_{model_index}.torch_model"

    # train the model
    print('Starting Training...')
    train_model = ModelTrainer(job_dir, device, model, criterion, optimizer, hooks, task)
    sucess = train_model(train_loader, num_epochs)
    return sucess


def set_random_seed(random_seed):
    # Set random seeds
    np.random.seed(random_seed)
    torch.random.manual_seed(random_seed)
    random.seed(random_seed)
    os.environ['PYTHONHASHSEED'] = str(random_seed)
    torch.manual_seed(random_seed)


def set_device(device):
    device = int(device) if device != 'cpu' else device
    device = torch.device(device if torch.cuda.is_available() else 'cpu')
    if device.type == 'cpu':
        print('Warning: running on CPU!')
    return device


def run_ensemble(job_dir, train_csv_path, valid_csv_path, config_file, num_epochs, device, random_seeds, overwrite):
    random_seeds = [int(seed) for seed in random_seeds.split()]

    if len(np.unique(random_seeds)) != len(random_seeds):
        raise ValueError("Duplicate random seeds were provided.")

    for idx, random_seed in enumerate(random_seeds, start=1):
        run_dir = os.path.join(job_dir, f"model_{idx}")
        run_job(run_dir, train_csv_path, valid_csv_path, config_file,
                num_epochs, device, random_seed, model_index=idx)

        if os.path.exists(run_dir):
            if overwrite:
                print('Run already exists, overwriting...')
            else:
                print('Run already exists, not overwriting...')
                continue
        run_job(run_dir, train_csv_path, valid_csv_path, config_file, num_epochs, device, random_seed)


def train():
    parser = argparse.ArgumentParser()
    parser.add_argument('--job-dir',
                        required=True,
                        type=str,
                        help='Directory for checkpoints, exports, and logs.')
    parser.add_argument('--config-file',
                        default=get_config(),
                        type=str,
                        help='Path to config file (defaults to packaged JSON).'
    )
    parser.add_argument('--train-csv-path',
                        required=True,
                        type=str,
                        help='Path to train csv file with paths of images, targets and masks.')
    parser.add_argument('--valid-csv-path',
                        required=True,
                        type=str,
                        help='Path to validation csv file with paths of images, targets and masks.')
    parser.add_argument('--num-epochs',
                        required=True,
                        type=int,
                        help='Number of epoch to train the model.')
    parser.add_argument('--device',
                        required=True,
                        type=str,
                        help='Device to use for computation')
    parser.add_argument('--random-seeds',
                        default="1 2 3 4 5",
                        type=str,
                        help='List of random seeds for training.')
    parser.add_argument('--overwrite',
                        default=False,
                        type=bool,
                        help='Whether to overwrite run if already exists')

    parse_args, unknown = parser.parse_known_args()

    run_ensemble(**parse_args.__dict__)


if __name__ == "__main__":
    train()



import os

import SimpleITK as sitk
import numpy as np
import pandas as pd


class LesionVolumeLocalisationMNI(object):
    def __init__(self, localisation_dir, native_space, atlas_label_map_path, brain_mask_path, roi_dictionary_csv,
                 label_map_name, write_registration_info):

        self.atlas_label_map = sitk.ReadImage(atlas_label_map_path)
        self.brain_mask = sitk.ReadImage(brain_mask_path, sitk.sitkInt8)
        roi_dictionary = pd.read_csv(roi_dictionary_csv)
        self.roi_dictionary = \
            {name: label for name, label in zip(roi_dictionary['ROIName'], roi_dictionary['ROIIndex'])}
        self.class_names = ['background', 'iph', 'eah', 'oedema', 'ivh']
        self.native_space = native_space
        self.localisation_dir = localisation_dir
        self.label_map_name = label_map_name
        self.write_registration_info = write_registration_info

    # Calculate total volume of a region or of the brain
    @staticmethod
    def calc_volume_ml(image):
        array = sitk.GetArrayFromImage(image)
        return np.sum(array) * np.prod(image.GetSpacing()) / 1000.

    def localise_lesion_volumes(self, label_map, atlas_label_map, brain_mask):
        # Creating a dictionary with 4 nested dictionaries (one per class). Inside those, the keys are the regions
        # labels and so far it is empty.
        localised_volumes = {class_name: {atlas_label_name: None for atlas_label_name in self.roi_dictionary}
                             for class_name in self.class_names}
        # Creating a dictionary to store the region volumes
        region_volumes = {atlas_label_name: None for atlas_label_name in self.roi_dictionary}
        for roi_name, roi_label in self.roi_dictionary.items():
            # if region is background (image_id 0)
            if roi_label == 0:
                continue
            # Create a mask for each region (each atlas label)
            region_mask = atlas_label_map == roi_label

            region_volumes[roi_name] = self.calc_volume_ml(sitk.Mask(region_mask, brain_mask))

            # Get a mask of the portion of each region occupied by the lesion
            masked_label_map = sitk.Mask(label_map, region_mask)
            for class_label, class_name in enumerate(self.class_names):
                # class_label = 0 -> background
                if class_label == 0:
                    continue

                masked_label_map_array = sitk.GetArrayFromImage(masked_label_map)
                masked_label_map_class = np.where(masked_label_map_array == class_label, 1, 0)
                masked_label_map_class = sitk.GetImageFromArray(masked_label_map_class)

                # Calculate the volume of each class type in the overlap between the region and lesion
                localised_volumes[class_name][roi_name] = self.calc_volume_ml(masked_label_map_class)
                # localised_volumes[class_name][roi_name] = self.calc_volume_ml(masked_label_map == class_label)

        return localised_volumes, region_volumes

    def __call__(self, aff_transform, data_index, image_id, label_map):
        # get the atlas_label_map, brain mask and label_map in the native or atlas space
        atlas_label_map = self.atlas_label_map
        brain_mask = self.brain_mask

        if self.native_space:
            atlas_label_map \
                = sitk.Resample(atlas_label_map, label_map, aff_transform.GetInverse(), sitk.sitkNearestNeighbor, 0)
            brain_mask = sitk.Resample(brain_mask, label_map, aff_transform.GetInverse(), sitk.sitkNearestNeighbor, 0)
        else:
            label_map = sitk.Resample(label_map, atlas_label_map, aff_transform, sitk.sitkNearestNeighbor, 0)

        localised_volumes, region_volumes = self.localise_lesion_volumes(label_map, atlas_label_map, brain_mask)
        for class_name in localised_volumes.keys():
            for roi_name in localised_volumes[class_name].keys():
                volume = localised_volumes[class_name][roi_name]
                if volume is not None:
                    data_index.loc[image_id, f'{self.label_map_name}_{class_name:s}_{roi_name:s}_ml'] = volume

        # add region volumes
        data_index.loc[image_id, f'Brain_volume_ml'] = self.calc_volume_ml(brain_mask)
        for roi_name in region_volumes.keys():
            volume = region_volumes[roi_name]
            if volume is not None:
                data_index.loc[image_id, f'{roi_name:s}_volume_ml'] = region_volumes[roi_name]

        if self.write_registration_info and self.native_space:
            atlas_native_space_path = os.path.join(self.localisation_dir, f'{str(image_id):s}_parc_atlas_native.nii.gz')
            sitk.WriteImage(atlas_label_map, atlas_native_space_path)
            brain_mask_native_space_path = os.path.join(self.localisation_dir,
                                                        f'{str(image_id):s}_brain_mask_native.nii.gz')
            sitk.WriteImage(brain_mask, brain_mask_native_space_path)
            data_index.loc[image_id, 'atlas_in_native_space'] = atlas_native_space_path
            data_index.loc[image_id, 'brain_mask_native_space'] = brain_mask_native_space_path
        return data_index



import hashlib
import pathlib
import requests
import sys
import tempfile

# allow direct file downloads, not HTML viewer pages
BASE_URL = "https://huggingface.co/sakshirathi360/blast-ct/resolve/main/"

def _sha256(path):
    h = hashlib.sha256()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(8192), b""):
            h.update(chunk)
    return h.hexdigest()

def _download(url, dest):
    tmp = pathlib.Path(tempfile.gettempdir()) / dest.name

    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        total = int(r.headers.get("Content-Length", 0))
        downloaded = 0

        with open(tmp, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                if not chunk:
                    break
                f.write(chunk)
                downloaded += len(chunk)

                # progress
                if total > 0:
                    done = int(50 * downloaded / total)
                    sys.stdout.write(f"\r[{'=' * done}{' ' * (50-done)}] {downloaded/1e6:.2f}MB/{total/1e6:.2f}MB")
                    sys.stdout.flush()

    print("\nDownload complete.")
    tmp.rename(dest)

def get_model_path(filename, expected_sha256=None):
    dest = pathlib.Path("~/.blastct/models").expanduser() / filename
    dest.parent.mkdir(parents=True, exist_ok=True)

    # already downloaded
    if dest.exists():
        if expected_sha256 is not None:
            if _sha256(dest) != expected_sha256:
                print("[blast_ct] Checksum mismatch, redownloading…")
                dest.unlink()
            else:
                return str(dest)
        else:
            return str(dest)

    print(f"[blast_ct] Downloading model: {filename}…")

    # 🔥 FIX: enforce binary download mode
    url = f"{BASE_URL}{filename}?download=1"

    _download(url, dest)

    if expected_sha256 is not None and _sha256(dest) != expected_sha256:
        dest.unlink()
        raise ValueError(f"[blast_ct] ERROR: SHA256 mismatch for {filename}")

    return str(dest)



import torch.nn
from torch.utils.data.dataloader import DataLoader

from blast_ct import models as models
from blast_ct.nifti import augmention
from blast_ct.nifti import patch_samplers
from blast_ct.nifti import transformation
from blast_ct.nifti.datasets import PatchWiseNiftiDataset, FullImageToOverlappingPatchesNiftiDataset, worker_init_fn
from blast_ct.nifti.savers import NiftiPatchSaver
from blast_ct.trainer import losses as losses
from blast_ct.trainer.hooks import TrainingEvaluator, ValidationEvaluator, ModelSaverHook, NaNLoss
from blast_ct.trainer.metrics import Loss
from blast_ct.trainer.metrics import SegmentationMetrics


def get_augmentation(augmentation_dict):
    return [getattr(augmention, name)(**kwargs) for name, kwargs in augmentation_dict.items()]


def get_transformation(transformation_dict):
    return [getattr(transformation, name)(**kwargs) for name, kwargs in transformation_dict.items()]


def get_train_loader(config, model, train_csv_path, use_cuda):
    input_patch_size = tuple(config['training']['input_patch_size'])
    output_patch_size = model.get_output_size(input_patch_size)

    transformation = get_transformation(config['data']['transformation'])
    augmentation = get_augmentation(config['training']['augmentation'])
    patch_augmentation = get_augmentation(config['training']['patch_augmentation'])

    sampler_type = list(config['training']['sampler'].keys())[0]
    config['training']['sampler'][sampler_type].update({'augmentation': patch_augmentation})
    sampler_class = getattr(patch_samplers, sampler_type)
    sampler = sampler_class(input_patch_size, output_patch_size, **config['training']['sampler'][sampler_type])

    sampling_mask = config['data']['sampling_mask'] if 'sampling_mask' in config['data'] else None
    sample_weight = config['data']['sample_weight'] if 'sample_weight' in config['data'] else None
    resolution = config['data']['resolution'] if 'resolution' in config['data'] else None

    train_set = PatchWiseNiftiDataset(patch_sampler=sampler,
                                      images_per_epoch=config['training']['images_per_epoch'],
                                      patches_per_image=config['training']['patches_per_image'],
                                      data_csv_path=train_csv_path,
                                      channels=config['data']['channels'],
                                      target=config['data']['target'],
                                      sampling_mask=sampling_mask,
                                      sample_weight=sample_weight,
                                      transformation=transformation,
                                      augmentation=augmentation,
                                      max_cases_in_memory=config['training']['max_cases_in_memory'],
                                      resolution=resolution)

    train_loader = DataLoader(train_set,
                              batch_size=config['training']['batch_size'],
                              num_workers=config['training']['num_workers'],
                              worker_init_fn=worker_init_fn,
                              pin_memory=True if use_cuda else False)
    return train_loader


def get_valid_loader(config, model, test_csv_path, use_cuda):
    if config['valid'] is None:
        return None
    input_patch_size = tuple(config['valid']['input_patch_size'])
    output_patch_size = model.get_output_size(input_patch_size)

    transformation = get_transformation(config['data']['transformation'])

    sampling_mask = config['data']['sampling_mask'] if 'sampling_mask' in config['data'] else None
    sample_weight = config['data']['sample_weight'] if 'sample_weight' in config['data'] else None
    resolution = config['data']['resolution'] if 'resolution' in config['data'] else None

    sampler_type = list(config['training']['sampler'].keys())[0]
    sampler_class = getattr(patch_samplers, sampler_type)
    sampler = sampler_class(input_patch_size, output_patch_size, **config['training']['sampler'][sampler_type])

    # set sequential to True to reduce disk access,
    valid_set = PatchWiseNiftiDataset(patch_sampler=sampler,
                                      images_per_epoch=config['valid']['images_per_epoch'],
                                      patches_per_image=config['valid']['patches_per_image'],
                                      data_csv_path=test_csv_path,
                                      channels=config['data']['channels'],
                                      target=config['data']['target'],
                                      sampling_mask=sampling_mask,
                                      sample_weight=sample_weight,
                                      transformation=transformation,
                                      max_cases_in_memory=config['training']['max_cases_in_memory'],
                                      sequential=True,
                                      resolution=resolution)

    valid_loader = DataLoader(valid_set,
                              batch_size=config['valid']['batch_size'],
                              num_workers=config['valid']['num_workers'],
                              worker_init_fn=worker_init_fn,
                              pin_memory=True if use_cuda else False)

    return valid_loader


def get_test_loader(config, model, test_csv_path, use_cuda):
    if config['test'] is None:
        return None
    input_patch_size = tuple(config['test']['input_patch_size'])
    output_patch_size = model.get_output_size(input_patch_size)
    transformation = get_transformation(config['data']['transformation'])

    sampling_mask = config['data']['sampling_mask'] if 'sampling_mask' in config['data'] else None
    resolution = config['data']['resolution'] if 'resolution' in config['data'] else None

    test_set = FullImageToOverlappingPatchesNiftiDataset(image_patch_shape=input_patch_size,
                                                         target_patch_shape=output_patch_size,
                                                         data_csv_path=test_csv_path,
                                                         channels=config['data']['channels'],
                                                         target=config['data']['target'],
                                                         sampling_mask=sampling_mask,
                                                         transformation=transformation,
                                                         resolution=resolution)

    test_loader = DataLoader(test_set,
                             batch_size=config['test']['batch_size'],
                             shuffle=False,
                             num_workers=config['test']['num_workers'],
                             worker_init_fn=worker_init_fn,
                             pin_memory=True if use_cuda else False)
    return test_loader


def get_training_hooks(job_dir, config, device, valid_loader, test_loader):
    hooks = []

    def get_metrics():
        return {'loss': Loss(device),
                'metrics': SegmentationMetrics(device, config['data']['class_names'])}

    hooks.append(TrainingEvaluator(job_dir + '/train', get_metrics()))
    hooks.append(NaNLoss())
    hooks.append(ModelSaverHook(config['valid']['eval_every'], config['valid']['keep_model_every']))

    if valid_loader is not None:
        hooks.append(ValidationEvaluator(job_dir + '/val', get_metrics(), valid_loader, config['valid']['eval_every']))

    # test only done at the end
    if test_loader is not None:
        extra_output_names = config['test']['extra_output_names'] if 'extra_output_names' in config['test'] else None
        saver = NiftiPatchSaver(job_dir + '/test', test_loader, extra_output_names=extra_output_names)
        hooks.append(ValidationEvaluator(job_dir + '/test',
                                         get_metrics(),
                                         test_loader,
                                         config['test']['eval_every'],
                                         saver))
    return hooks


def get_model(config):
    model_type = list(config['model'].keys())[0]
    model_class = getattr(models, model_type)
    model = model_class(input_channels=config['data']['input_channels'],
                        num_classes=config['data']['num_classes'],
                        **config['model'][model_type])
    return model


def get_loss(config):
    loss_type = list(config['loss'].keys())[0]
    loss_class = getattr(losses, loss_type)
    loss = loss_class(**config['loss'][loss_type])
    return loss


def get_optimizer(config, model):
    optimizer_type = list(config['optimizer'].keys())[0]
    optimizer_class = getattr(torch.optim, optimizer_type)
    optimizer = optimizer_class(model.parameters(), **config['optimizer'][optimizer_type])
    scheduler_type = list(config['scheduler'].keys())[0]
    scheduler_class = getattr(torch.optim.lr_scheduler, scheduler_type)
    scheduler = scheduler_class(optimizer, **config['scheduler'][scheduler_type])
    return scheduler



import os
from pathlib import Path

def get_config():
    """
    Returns the path to the packaged config.json.
    Assumes config is installed inside blast_ct/data/.
    """
    install_dir = os.path.dirname(os.path.realpath(__file__))
    config_path = Path(install_dir).parent / "data" / "config.json"
    return str(config_path)



import time
import os
import re
import torch
from abc import ABC
import numpy as np
from blast_ct.trainer.metrics import to_np_cpu, Metric
from torch.utils.tensorboard import SummaryWriter


class Hook(ABC):
    def __init__(self):
        self.model_trainer = None

    def attach_hook(self, model_trainer):
        self.model_trainer = model_trainer

    def before_epoch(self):
        pass

    def after_batch(self):
        pass

    def after_epoch(self):
        pass


class Evaluator(object):
    def __init__(self, log_dir, metrics):
        for metric in metrics.values():
            assert isinstance(metric, Metric)
        self.metrics = metrics
        self.writer = SummaryWriter(log_dir=log_dir)

    def increment_state(self, model_state):
        for metric in self.metrics.values():
            metric.increment(model_state)

    def calculate_and_reset_metrics(self):
        for metric in self.metrics.values():
            metric.save_and_reset()

    def report(self):
        message = ''
        for name, metric in self.metrics.items():
            message += metric.report()
        return message

    def log_to_tensorboard(self, epoch):
        for name, metric in self.metrics.items():
            metric.log_to_tensorboard(epoch, self.writer, name)


class TrainingEvaluator(Hook, Evaluator):
    def __init__(self, log_dir, metrics):
        Hook.__init__(self)
        Evaluator.__init__(self, log_dir, metrics)
        self.time = 0

    def before_epoch(self):
        self.time = time.time()

    def after_batch(self):
        self.increment_state(self.model_trainer.current_state)

    def after_epoch(self):
        self.calculate_and_reset_metrics()
        epoch = self.model_trainer.current_state['epoch']
        num_epochs = self.model_trainer.current_state['num_epochs']
        elapsed_time = time.time() - self.time
        message = f'Training epoch {epoch:d}/{num_epochs - 1:d} completed in {elapsed_time:.0f}s\n' + self.report()
        self.model_trainer.logger.info(message)
        self.log_to_tensorboard(epoch)


class ValidationEvaluator(Hook, Evaluator):
    def __init__(self, log_dir, metrics, dataloader, eval_every, saver=None):
        Hook.__init__(self)
        Evaluator.__init__(self, log_dir, metrics)
        self.dataloader = dataloader
        self.eval_every = eval_every
        self.saver = saver

    def perform_evaluation(self):
        epoch = self.model_trainer.current_state['epoch']
        for state in self.model_trainer.step(epoch, self.dataloader, is_training=False):
            self.increment_state(state)
            if self.saver is not None:
                message = self.saver(state)
                if message is not None:
                    self.model_trainer.logger.info(message)

    def before_epoch(self):
        if self.saver is not None:
            self.saver.reset()

    def after_epoch(self):
        epoch = self.model_trainer.current_state['epoch']
        num_epochs = self.model_trainer.current_state['num_epochs']
        is_last_epoch = epoch == num_epochs - 1
        if not (epoch % self.eval_every == 0 or is_last_epoch) or epoch == 0:
            return

        start_time = time.time()
        self.perform_evaluation()
        self.calculate_and_reset_metrics()
        elapsed_time = time.time() - start_time
        message = f'Validation epoch {epoch:d}/{num_epochs - 1:d} completed in {elapsed_time:.0f}s\n' + self.report()
        self.model_trainer.logger.info(message)
        self.log_to_tensorboard(epoch)


class NaNLoss(Hook):
    def __init__(self):
        super().__init__()

    def after_epoch(self):
        if np.isnan(to_np_cpu(self.model_trainer.current_state['loss'])):
            self.model_trainer.stop_signal = True
            self.model_trainer.logger.error('Found NaN loss during training, stopping gracefully.')


class ModelSaverHook(Hook):
    def __init__(self, save_every=100, keep_model_every=100, filename=None):
        super().__init__()
        self.save_every = save_every
        self.keep_model_every = keep_model_every
        self.filename = filename

    @staticmethod
    def save_model_to_disk(model, model_path):
        os.makedirs(os.path.dirname(model_path), exist_ok=True)
        torch.save(model, model_path)

    def after_epoch(self):
        epoch = self.model_trainer.current_state['epoch']
        num_epochs = self.model_trainer.current_state['num_epochs']
        is_last_epoch = epoch == num_epochs - 1

        if not (epoch % self.save_every == 0 or is_last_epoch) or epoch == 0:
            return

        saved_model_dir = os.path.join(self.model_trainer.job_dir, 'saved_models')

        # 🔥 Use custom filename if set, else fallback to epoch-based
        filename = self.filename or f"model_{epoch}.torch_model"
        model_path = os.path.join(saved_model_dir, filename)

        self.save_model_to_disk(self.model_trainer.model.state_dict(), model_path)

        # 🔥 If filename is fixed, do NOT delete models based on epoch
        if self.filename:
            return

        # delete old models (epoch mode only)
        for file in os.listdir(saved_model_dir):
            matches = re.findall('(?<=model_)\d+', file)
            for match in matches:
                previous_epoch = int(match)
                if (previous_epoch % self.keep_model_every != 0 or previous_epoch == 0) and previous_epoch != epoch:
                    os.remove(os.path.join(saved_model_dir, file))

        if is_last_epoch and not self.filename:
            last_path = os.path.join(saved_model_dir, 'model_last.torch_model')
            self.save_model_to_disk(self.model_trainer.model.state_dict(), last_path)



import numpy as np
from scipy.ndimage.filters import gaussian_filter


# All augmentations can't alter state after __init__ because of dataset class when num_workers > 0


class RandomAugmentation(object):
    """
    Abstract class for random patch augmentation, patch augmentation also works on full images
    __call__: When called a Augmentation should return an image and target and mask with the same shape
    as the input.
    """

    def __init__(self, prob):
        self.prob = prob

    def __call__(self, image, target=None, mask=None):
        if np.random.choice((True, False), p=(self.prob, 1. - self.prob)):
            return self.augment(image, target, mask)
        else:
            return image, target, mask

    def augment(self, image, target, mask):
        raise NotImplementedError


class RandomPatchRotation(RandomAugmentation):
    def __init__(self, prob, allowed_planes, rotations=(1, 2, 3)):
        super().__init__(prob)
        self.allowed_planes = allowed_planes
        self.rotations = rotations

    def augment(self, image, target, mask):
        k = np.random.choice(self.rotations, len(self.allowed_planes))
        for i, axes in enumerate(self.allowed_planes):
            axes = np.random.choice(axes, 2, replace=False)  # direction of rotation
            image = np.rot90(image, k=k[i], axes=tuple(a + 1 for a in axes))
            target = np.rot90(target, k=k[i], axes=axes) if target is not None else None
            mask = np.rot90(mask, k=k[i], axes=axes) if mask is not None else None

        return image, target, mask


class RandomPatchFlip(RandomAugmentation):
    def __init__(self, prob, allowed_axis):
        super().__init__(prob)
        self.allowed_axes = allowed_axis

    def augment(self, image, target, mask):
        for axis in self.allowed_axes:
            image = np.flip(image, axis=axis + 1)
            target = np.flip(target, axis=axis) if target is not None else None
            mask = np.flip(mask, axis=axis) if mask is not None else None

        return image, target, mask


class RandomHistogramDeformation(RandomAugmentation):
    def __init__(self, prob, shift_std=0.05, scale_std=0.01, allow_mirror=False):
        super().__init__(prob)
        self.shift_std = shift_std
        self.scale_std = scale_std
        self.allow_mirror = allow_mirror

    def augment(self, image, target, mask):
        num_channels = image.shape[0]
        shift = np.random.uniform(0, self.shift_std, num_channels)
        scale = np.random.normal(1, self.scale_std, num_channels)
        if self.allow_mirror:
            scale *= np.random.choice((-1, 1))
        image = (image.T * scale).T
        image = (image.T + shift).T
        return image, target, mask


class RandomGammaCorrection(RandomAugmentation):
    def __init__(self, prob, range_min=-1., range_max=1., gamma_std=.1):
        super().__init__(prob)
        self.range_min = range_min
        self.range_max = range_max
        self.gamma_std = gamma_std

    def augment(self, image, target, mask):
        num_channels = image.shape[0]
        # gamma correction must be performed in the range of 0 to 1
        image = (image - self.range_min) / (self.range_max - self.range_min)
        gamma = np.random.normal(1, self.gamma_std, num_channels)
        image = np.power(image.T, gamma).T
        image = image * (self.range_max - self.range_min) + self.range_min
        return image, target, mask


class RandomElasticDeformation(RandomAugmentation):
    """
    alpha: The amplitude of the noise;
    prob: Probability of deformation occurring
    noise_shape: Shape of the deformation field from which to sample patches from (must be larger than input_shape)
    num_maps Number of different noise maps to generate
    """

    def __init__(self, prob, alpha, noise_shape, num_maps=3):
        super().__init__(prob)
        self.alpha = alpha
        self.num_maps = num_maps
        self.noise_shape = noise_shape
        self.deformation_fields = [np.round(self.get_1d_displacement_field(self.noise_shape)).astype(np.int32)
                                   for _ in range(self.num_maps)]
        self.patch_shape = None
        self.grid = None

    def get_1d_displacement_field(self, shape):
        raise NotImplementedError

    def get_displacement_field(self, patch_shape):
        dx = [self.deformation_fields[i] for i in np.random.choice(len(self.deformation_fields), 3)]
        starts = [[np.random.randint(s - ps + 1) for s, ps in zip(self.noise_shape, patch_shape)] for _ in range(3)]
        slices = [tuple(slice(s, s + ps, 1) for s, ps in zip(start, patch_shape)) for start in starts]
        return [dx[i][slices[i]] for i in range(3)]

    def augment(self, image, target, mask):

        shape = image.shape[1:]
        if shape != self.patch_shape:
            self.grid = [g.astype(np.int32) for g in np.meshgrid(*[np.arange(s) for s in shape], indexing='ij')]
            self.patch_shape = shape

        dx = self.get_displacement_field(shape)

        indices = sum(np.clip((x_i + dx_i), a_min=0, a_max=s_i - 1).reshape(-1, 1) *
                      np.prod(shape[(i + 1):]).astype(np.int32)
                      for i, (x_i, dx_i, s_i) in enumerate(zip(self.grid, dx, shape)))
        try:
            target = target.ravel()[indices].reshape(shape) if target is not None else None
            mask = mask.ravel()[indices].reshape(shape) if mask is not None else None
            image = np.stack([channel.ravel()[indices].reshape(shape) for channel in image])
        except IndexError:
            return image, target, mask

        return image, target, mask


class RandomElasticDeformationSimard2003(RandomElasticDeformation):
    """Elastic deformation of images as described in [Simard2003]_.
    .. [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
       Convolutional Neural Networks applied to Visual Document Analysis", in
       Proc. of the International Conference on Document Analysis and
       Recognition, 2003.
    """

    def __init__(self, prob, sigma, alpha, noise_shape, num_maps=3):
        self.sigma = sigma
        super().__init__(prob, alpha, noise_shape, num_maps)

    def get_1d_displacement_field(self, shape):
        dx = np.random.rand(*shape) * 2 - 1
        dx = gaussian_filter(dx, self.sigma, mode="nearest") * self.alpha
        return dx


class RandomElasticDeformationCoarse(RandomElasticDeformationSimard2003):
    def __init__(self, prob, sigma, coarseness, alpha, noise_shape, num_maps=3):
        self.coarseness = coarseness
        super().__init__(prob, sigma, alpha, noise_shape, num_maps)

    def get_1d_displacement_field(self, shape):
        coarse_shape = tuple(s // c + bool(s % c) for s, c in zip(shape, self.coarseness))
        dx = np.random.rand(*coarse_shape) * 2 - 1
        dx = np.kron(dx, np.ones(shape=self.coarseness))
        dx = gaussian_filter(dx, self.sigma, mode="nearest") * self.alpha
        dx = dx[tuple(slice(0, s, 1) for s in shape)]
        return dx


class RandomElasticDeformationCoarsePerlinNoise(RandomElasticDeformation):

    def __init__(self, prob, period, alpha, noise_shape, num_maps=3):
        self.period = period
        super().__init__(prob, alpha, noise_shape, num_maps)

    def get_1d_displacement_field(self, shape):
        new_shape = tuple((s // c + bool(s % c) + bool(c == 1)) * c for s, c in zip(shape, self.period))
        dx = self.generate_fractal_noise_3d(new_shape, self.period)
        dx = dx[tuple(slice(0, s, 1) for s in shape)] * self.alpha
        return dx

    # https://github.com/pvigier/perlin-numpy
    def generate_fractal_noise_3d(self, shape, res, octaves=1, persistence=0.5):
        noise = np.zeros(shape)
        frequency = 1
        amplitude = 1
        for _ in range(octaves):
            noise += amplitude * self.generate_perlin_noise_3d(shape, (
                frequency * res[0], frequency * res[1], frequency * res[2]))
            frequency *= 2
            amplitude *= persistence
        return noise

    @staticmethod
    def generate_perlin_noise_3d(shape, res):
        def f(t):
            return 6 * t ** 5 - 15 * t ** 4 + 10 * t ** 3

        delta = (res[0] / shape[0], res[1] / shape[1], res[2] / shape[2])
        d = (shape[0] // res[0], shape[1] // res[1], shape[2] // res[2])
        grid = np.mgrid[0:res[0]:delta[0], 0:res[1]:delta[1], 0:res[2]:delta[2]]
        grid = grid.transpose(1, 2, 3, 0) % 1
        # Gradients
        random_state = np.random.RandomState(843)
        theta = 2 * np.pi * random_state.rand(res[0] + 1, res[1] + 1, res[2] + 1)
        phi = 2 * np.pi * random_state.rand(res[0] + 1, res[1] + 1, res[2] + 1)
        gradients = np.stack((np.sin(phi) * np.cos(theta), np.sin(phi) * np.sin(theta), np.cos(phi)), axis=3)

        g000 = gradients[0:-1, 0:-1, 0:-1].repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)
        g100 = gradients[1:, 0:-1, 0:-1].repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)
        g010 = gradients[0:-1, 1:, 0:-1].repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)
        g110 = gradients[1:, 1:, 0:-1].repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)
        g001 = gradients[0:-1, 0:-1, 1:].repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)
        g101 = gradients[1:, 0:-1, 1:].repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)
        g011 = gradients[0:-1, 1:, 1:].repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)
        g111 = gradients[1:, 1:, 1:].repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)

        # Ramps
        n000 = np.sum(np.stack((grid[:, :, :, 0], grid[:, :, :, 1], grid[:, :, :, 2]), axis=3) * g000, 3)
        n100 = np.sum(np.stack((grid[:, :, :, 0] - 1, grid[:, :, :, 1], grid[:, :, :, 2]), axis=3) * g100, 3)
        n010 = np.sum(np.stack((grid[:, :, :, 0], grid[:, :, :, 1] - 1, grid[:, :, :, 2]), axis=3) * g010, 3)
        n110 = np.sum(np.stack((grid[:, :, :, 0] - 1, grid[:, :, :, 1] - 1, grid[:, :, :, 2]), axis=3) * g110, 3)
        n001 = np.sum(np.stack((grid[:, :, :, 0], grid[:, :, :, 1], grid[:, :, :, 2] - 1), axis=3) * g001, 3)
        n101 = np.sum(np.stack((grid[:, :, :, 0] - 1, grid[:, :, :, 1], grid[:, :, :, 2] - 1), axis=3) * g101, 3)
        n011 = np.sum(np.stack((grid[:, :, :, 0], grid[:, :, :, 1] - 1, grid[:, :, :, 2] - 1), axis=3) * g011, 3)
        n111 = np.sum(np.stack((grid[:, :, :, 0] - 1, grid[:, :, :, 1] - 1, grid[:, :, :, 2] - 1), axis=3) * g111, 3)
        # Interpolation
        t = f(grid)
        n00 = n000 * (1 - t[:, :, :, 0]) + t[:, :, :, 0] * n100
        n10 = n010 * (1 - t[:, :, :, 0]) + t[:, :, :, 0] * n110
        n01 = n001 * (1 - t[:, :, :, 0]) + t[:, :, :, 0] * n101
        n11 = n011 * (1 - t[:, :, :, 0]) + t[:, :, :, 0] * n111
        n0 = (1 - t[:, :, :, 1]) * n00 + t[:, :, :, 1] * n10
        n1 = (1 - t[:, :, :, 1]) * n01 + t[:, :, :, 1] * n11

        return (1 - t[:, :, :, 2]) * n0 + t[:, :, :, 2] * n1



import torch
import torch.nn as nn

from blast_ct.models.base import BiomedicalBlock, DownSample, UpSample, PreActBlock, crop_center

SCALE_FACTORS = ((5, 5, 5), (3, 3, 3), (1, 1, 1))
FEATURE_MAPS = (30, 30, 40, 40, 40, 40, 50, 50)
FULLY_CONNECTED = (250, 250)
DROPOUT = (.0, .5, .5)


class Path(BiomedicalBlock):
    def __init__(self, scale_factor, input_channels, feature_maps):
        super().__init__()
        self.layers = list()
        self.scale_factor = tuple(scale_factor)

        self.layers.append(DownSample(self.scale_factor))
        for i, feature_map in enumerate(feature_maps):
            in_channels = feature_maps[i - 1] if i > 0 else input_channels
            self.layers.append(PreActBlock(in_channels, feature_map))
        self.layers.append(UpSample(self.scale_factor))

        self.path = nn.Sequential(*self.layers)

    def forward(self, x, output_size):
        input_size = self.calculate_input_size(output_size)
        out = crop_center(x, input_size)
        out = self.path(out)
        out = crop_center(out, output_size)
        return out


class DeepMedic(BiomedicalBlock):
    def __init__(self,
                 input_channels,
                 num_classes,
                 scale_factors=SCALE_FACTORS,
                 feature_maps=FEATURE_MAPS,
                 fully_connected=FULLY_CONNECTED,
                 dropout=DROPOUT):

        super().__init__()
        # assert all scale factors are equal or less than the next one
        assert all([all(l[i] >= l[i + 1] for i in range(len(l) - 1)) for l in [i for i in list(zip(*scale_factors))]])
        self.scale_factors = tuple(scale_factors)
        self.feature_maps = tuple(feature_maps)
        self.output_size = None

        self.paths = []
        for i, scale_factor in enumerate(scale_factors):
            path = Path(scale_factor, input_channels, feature_maps)
            self.paths.append(path)
            self.add_module('path' + str(scale_factor) + str(i), path)

        assert len(fully_connected) + 1 == len(dropout)
        fms = []
        channels = (feature_maps[-1] * len(self.paths),) + tuple(fully_connected) + (num_classes, )
        for i in range(len(channels[:-1])):
            fms.append(PreActBlock(channels[i], channels[i + 1], kernel_size=(1, 1, 1), dropout_prob=dropout[i]))

        self.fully_connected = nn.Sequential(*fms)

        # to calculate sizes
        self.layers = self.paths[0].layers

    def forward(self, image, **kwargs):
        input_size = tuple(image.shape[2:])
        output_size = self.get_output_size(input_size)

        activations = []
        for i, path in enumerate(self.paths):
            out = path(image, output_size)
            activations.append(out)

        out = torch.cat(tuple(activations), dim=1)
        out = self.fully_connected(out)
        return out, {}



import SimpleITK as sitk
import numpy as np


def get_physical_size(image):
    return [(sz - 1) * spc for sz, spc in zip(image.GetSize(), image.GetSpacing())]


def get_size_from_spacing(spacing, physical_size):
    return [int(round(phys_sz / spc + 1)) for spc, phys_sz in zip(spacing, physical_size)]


# Create reference image with origin at 0
def get_reference_image(image, new_spacing):
    physical_size = get_physical_size(image)
    size = get_size_from_spacing(new_spacing, physical_size)
    reference_image = sitk.Image(size, image.GetPixelID())
    reference_image.SetOrigin(image.GetOrigin())
    reference_image.SetDirection(image.GetDirection())
    reference_image.SetSpacing(new_spacing)
    return reference_image


def rescale(spacing, image: sitk.Image, is_discrete=False):
    reference_image = get_reference_image(image, spacing)
    mode = sitk.sitkNearestNeighbor if is_discrete else sitk.sitkLinear
    default_value = 0 if is_discrete else float(np.min(sitk.GetArrayViewFromImage(image)))
    return sitk.Resample(image, reference_image, sitk.Transform(), mode, default_value)


def sitk_is_vector(sitk_image):
    return sitk_image.GetPixelID() in range(12, 22)


def sitk_to_numpy(sitk_image, dtype=np.float32):
    if sitk_image is None:
        return None
    array = sitk.GetArrayFromImage(sitk_image).astype(dtype)
    if sitk_is_vector(sitk_image):
        if sitk_image.GetNumberOfComponentsPerPixel() == 1:
            array = np.expand_dims(array, axis=0)
        else:
            array = np.transpose(array, (-1,) + tuple(range(len(array.shape) - 1)))
    return array


def create_reference_reoriented_image(image):
    dir = np.array(image.GetDirection()).reshape(len(image.GetSize()), -1)
    ind = np.argmax(np.abs(dir), axis=0)
    new_size = np.array(image.GetSize())[ind]
    new_spacing = np.array(image.GetSpacing())[ind]
    new_extent = new_size * new_spacing
    new_dir = dir[:, ind]

    flip = np.diag(new_dir) < 0
    flip_diag = flip * -1
    flip_diag[flip_diag == 0] = 1
    flip_mat = np.diag(flip_diag)

    new_origin = np.array(image.GetOrigin()) + np.matmul(new_dir, (new_extent * flip))
    new_dir = np.matmul(new_dir, flip_mat)

    reference = sitk.Image(new_size.tolist(), image.GetPixelIDValue())
    reference.SetSpacing(new_spacing.tolist())
    reference.SetOrigin(new_origin.tolist())
    try:
        reference.SetDirection(new_dir.flatten().tolist())
    except RuntimeError:
        print('Could not reorient image due to singular direction matrix, proceeding with image not reoriented!')
        return image
    return reference


def reorient_image(image, is_discrete):
    """Reorients an image to standard radiology view."""
    default_value = 0 if is_discrete else float(np.min(sitk.GetArrayViewFromImage(image)))
    reference = create_reference_reoriented_image(image)
    resample = sitk.ResampleImageFilter()
    resample.SetReferenceImage(reference)
    resample.SetDefaultPixelValue(default_value)
    resample.SetTransform(sitk.Transform())
    if is_discrete:
        resample.SetInterpolator(sitk.sitkNearestNeighbor)
    else:
        resample.SetInterpolator(sitk.sitkLinear)
    return resample.Execute(image)



import torch
import torch.nn.functional as F
import time
from blast_ct.trainer.logger import get_logger


def predict_exclusive(logits):
    prob = F.softmax(logits, dim=1)
    _, pred = torch.max(logits, dim=1)
    return prob, pred


def predict_multi_target(logits):
    prob = torch.sigmoid(logits)
    pred = torch.round(prob)
    return prob, pred


def predict_regression(logits):
    return logits, logits


task_predict_fn_dict = {'segmentation': predict_exclusive,
                        'classification': predict_exclusive,
                        'multi_target_classification': predict_multi_target,
                        'regression': predict_regression}


class ModelTrainer(object):
    def __init__(self, job_dir, device, model, criterion, lr_scheduler, hooks, task):
        assert task in task_predict_fn_dict.keys()
        self.job_dir = job_dir
        self.device = device
        self.model = model
        self.criterion = criterion
        self.lr_scheduler = lr_scheduler
        self.hooks = hooks
        self.predict_fn = task_predict_fn_dict[task]
        self.logger = get_logger(job_dir)
        self.stop_signal = False

    def step(self, epoch, dataloader, is_training=True):
        self.model.train() if is_training else self.model.eval()
        for inputs in dataloader:
            inputs = {key: value.to(self.device) for key, value in inputs.items()}
            self.lr_scheduler.optimizer.zero_grad()
            with torch.set_grad_enabled(is_training):
                logits, state = self.model(**inputs)
                state.update(inputs)
                loss = self.criterion(logits, **state)
                if is_training:
                    loss.backward()
                    self.lr_scheduler.optimizer.step()
            prob, pred = self.predict_fn(logits)
            state.update({'epoch': epoch, 'loss': loss, 'logits': logits, 'prob': prob, 'pred': pred})
            yield state

    def _run_epoch(self, epoch, dataloader):
        [hook.before_epoch() for hook in self.hooks]
        for state in self.step(epoch, dataloader):
            self.current_state.update(state)
            [hook.after_batch() for hook in self.hooks]
        [hook.after_epoch() for hook in self.hooks]
        self.lr_scheduler.step()
        return

    def __call__(self, dataloader, num_epochs):
        start_training = time.time()
        self.model.to(self.device)
        [hook.attach_hook(self) for hook in self.hooks]
        self.current_state = {'num_epochs': num_epochs}

        for epoch in range(num_epochs):
            if self.stop_signal:
                break
            self._run_epoch(epoch, dataloader)

        time_elapsed = time.time() - start_training
        self.logger.info(f'Training completed in {time_elapsed // 3600:.0f}h {time_elapsed % 3600 // 60:.0f}m '
                         f'{time_elapsed % 3600 % 60:.0f}s')
        return False if self.stop_signal else True
